from typing import Type

from wai.annotations.core.domain import DomainSpecifier

from ._LabelDistributionOutputFormatSpecifier import LabelDistributionOutputFormatSpecifier


class LabelDistributionODOutputFormatSpecifier(LabelDistributionOutputFormatSpecifier):
    """
    Specifier for label-dist in the object-detection domain.
    """
    @classmethod
    def domain(cls) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        return ImageObjectDetectionDomainSpecifier



from abc import ABC
from typing import Type, Tuple

from wai.annotations.core.component import Component
from wai.annotations.core.specifier import SinkStageSpecifier


class LabelDistributionOutputFormatSpecifier(SinkStageSpecifier, ABC):
    """
    Base specifier for the label-dist in each known domain.
    """
    @classmethod
    def description(cls) -> str:
        return "Generates a label distribution."

    @classmethod
    def components(cls) -> Tuple[Type[Component], ...]:
        from wai.annotations.imgstats.format.labeldist.component import LabelDistribution
        return LabelDistribution,



import io
from PIL import Image, ImageDraw

from wai.annotations.core.component import SinkComponent
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance

from wai.common.cli.options import TypedOption, FlagOption


class AnnotationOverlay(
    SinkComponent[ImageObjectDetectionInstance]
):

    color: str = TypedOption(
        "-c", "--color",
        type=str,
        default="255,0,0,64",
        help="the color to use for drawing the shapes as RGBA byte-quadruplet, e.g.: 255,0,0,64"
    )

    background_color: str = TypedOption(
        "-b", "--background-color",
        type=str,
        default="255,255,255,255",
        help="the color to use for the background as RGBA byte-quadruplet, e.g.: 255,255,255,255"
    )

    output_file: str = TypedOption(
        "-o", "--output",
        type=str,
        default="./overlay.png",
        help="the PNG image to write the generated overlay to"
    )

    def output_overlay(self):
        """
        Outputs the overlay image.
        """
        if hasattr(self, "_overlay"):
            self._overlay.save(self.output_file)
        else:
            print("No overlay generated!")

    def consume_element(self, element: ImageObjectDetectionInstance):
        """
        Consumes instances.
        """
        img = Image.open(io.BytesIO(element.data.data))

        if not hasattr(self, "_overlay"):
            # initialize overlay
            self._color = tuple([int(x) for x in self.color.split(",")])
            self._background_color = tuple([int(x) for x in self.background_color.split(",")])
            self._overlay = Image.new('RGBA', img.size, self._background_color)
        else:
            # do we have to make the overlay larger?
            if (img.size[0] > self._overlay.size[0]) or (img.size[1] > self._overlay.size[1]):
                new_size = (max(img.size[0], self._overlay.size[0]), max(img.size[1], self._overlay.size[1]))
                tmp = Image.new('RGBA', new_size, self._background_color)
                tmp.paste(self._overlay, (0, 0))
                self._overlay = tmp

        for lobj in element.annotations:
            points = []
            if lobj.has_polygon():
                poly_x = lobj.get_polygon_x()
                poly_y = lobj.get_polygon_y()
                for x, y in zip(poly_x, poly_y):
                    points.append((x, y))
            else:
                rect = lobj.get_rectangle()
                points.append((rect.left(), rect.top()))
                points.append((rect.right(), rect.top()))
                points.append((rect.right(), rect.bottom()))
                points.append((rect.left(), rect.bottom()))

            draw = ImageDraw.Draw(self._overlay)
            draw.polygon(tuple(points), outline=self._color)

    def finish(self):
        self.output_overlay()



from ._LabelDistributionICOutputFormatSpecifier import LabelDistributionICOutputFormatSpecifier
from ._LabelDistributionISOutputFormatSpecifier import LabelDistributionISOutputFormatSpecifier
from ._LabelDistributionODOutputFormatSpecifier import LabelDistributionODOutputFormatSpecifier



import csv
import json
import sys
from collections import OrderedDict
import numpy as np

from wai.annotations.core.component import SinkComponent
from wai.annotations.domain.image import ImageInstance
from wai.annotations.domain.image.classification import ImageClassificationInstance
from wai.annotations.domain.image.object_detection import ImageObjectDetectionInstance
from wai.annotations.domain.image.segmentation import ImageSegmentationInstance

from wai.common.cli.options import TypedOption, FlagOption

OUTPUT_FORMAT_TEXT = "text"
OUTPUT_FORMAT_CSV = "csv"
OUTPUT_FORMAT_JSON = "json"
OUTPUT_FORMATS = [
    OUTPUT_FORMAT_CSV,
    OUTPUT_FORMAT_JSON,
]


class LabelDistribution(
    SinkComponent[ImageInstance]
):
    output_format: str = TypedOption(
        "-f", "--format",
        type=str,
        default=OUTPUT_FORMAT_TEXT,
        help="the format to use for the output, available modes: %s" % ", ".join(OUTPUT_FORMATS)
    )

    output_file: str = TypedOption(
        "-o", "--output",
        type=str,
        default="",
        help="the file to write the statistics to; uses stdout if omitted"
    )

    percentages: bool = FlagOption(
        "-p", "--percentages",
        help="whether to output percentages instead of counts."
    )

    def init_labels(self):
        """
        Initializes the labels.
        """
        if not hasattr(self, "_labels"):
            self._labels = dict()

    def add_label(self, label):
        """
        Increments the count for the label.
        """
        self.init_labels()

        if label not in self._labels:
            self._labels[label] = 0

        self._labels[label] = self._labels[label] + 1

    def output_text(self, dist, keys, use_stdout):
        """
        Outputs the label distribution in simple textual format.

        :param dist: the distribution dictionary
        :type dist: dict
        :param keys: they sorted keys
        :type keys: list
        :param use_stdout: whether to use stdout or the file
        :type use_stdout: bool
        """
        if use_stdout:
            for k in keys:
                if self.percentages:
                    print("%s: %f" % (k, dist[k]))
                else:
                    print("%s: %d" % (k, dist[k]))
        else:
            with open(self.output_file, "w") as f:
                for k in keys:
                    if self.percentages:
                        f.write("%s: %f" % (k, dist[k]))
                    else:
                        f.write("%s: %d" % (k, dist[k]))
                    f.write("\n")

    def output_csv(self, dist, keys, use_stdout):
        """
        Outputs the label distribution in CSV format.

        :param dist: the distribution dictionary
        :type dist: dict
        :param keys: they sorted keys
        :type keys: list
        :param use_stdout: whether to use stdout or the file
        :type use_stdout: bool
        """
        if use_stdout:
            writer = csv.writer(sys.stdout)
            f = None
        else:
            f = open(self.output_file, "w")
            writer = csv.writer(f)

        writer.writerow(["Label", "Percent" if self.percentages else "Count"])
        for k in keys:
            writer.writerow([k, dist[k]])

        if f is not None:
            f.close()

    def output_json(self, dist, use_stdout):
        """
        Outputs the label distribution in json format.

        :param dist: the distribution dictionary
        :type dist: dict
        :param keys: they sorted keys
        :type keys: list
        :param use_stdout: whether to use stdout or the file
        :type use_stdout: bool
        """
        if use_stdout:
            print(json.dumps(dist, indent=2))
        else:
            with open(self.output_file, "w") as f:
                json.dump(dist, f, indent=2)

    def output_label_distribution(self):
        """
        Outputs the distribution.
        """
        keys = list(self._labels.keys())
        sorted(keys)
        dist = OrderedDict()
        for k in keys:
            dist[k] = self._labels[k]

        if self.percentages:
            total = 0
            for k in dist:
                total += dist[k]
            for k in dist:
                dist[k] = dist[k] / total * 100.0

        use_stdout = len(self.output_file) == 0

        if self.output_format == "text":
            self.output_text(dist, keys, use_stdout)
        elif self.output_format == "csv":
            self.output_csv(dist, keys, use_stdout)
        elif self.output_format == "json":
            self.output_json(dist, use_stdout)
        else:
            raise Exception("Unhandled output format: %s" % self.output_format)

    def consume_element(self, element: ImageInstance):
        """
        Consumes instances by discarding them.
        """
        if isinstance(element, ImageClassificationInstance):
            self.add_label(element.annotations.label)
        elif isinstance(element, ImageObjectDetectionInstance):
            for obj in element.annotations:
                if "type" in obj.metadata:
                    self.add_label(obj.metadata["type"])
        elif isinstance(element, ImageSegmentationInstance):
            unique = np.unique(element.annotations.indices)
            for index in unique:
                # skip background
                if index == 0:
                    continue
                self.add_label(element.annotations.labels[index - 1])

    def finish(self):
        self.init_labels()
        self.output_label_distribution()



from typing import Type

from wai.annotations.core.domain import DomainSpecifier

from ._LabelDistributionOutputFormatSpecifier import LabelDistributionOutputFormatSpecifier


class LabelDistributionICOutputFormatSpecifier(LabelDistributionOutputFormatSpecifier):
    """
    Specifier for the label-dist in the image-classification domain.
    """
    @classmethod
    def domain(cls) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.classification import ImageClassificationDomainSpecifier
        return ImageClassificationDomainSpecifier



from setuptools import setup, find_namespace_packages


def _read(filename: str) -> str:
    """
    Reads in the content of the file.

    :param filename:    The file to read.
    :return:            The file content.
    """
    with open(filename, "r") as file:
        return file.read()


setup(
    name="wai.annotations.imgstats",
    description="Various sinks for generating image dataset statistics.",
    long_description=f"{_read('DESCRIPTION.rst')}\n"
                     f"{_read('CHANGES.rst')}",
    url="https://github.com/waikato-ufdl/wai-annotations-imgstats",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: Apache Software License',
        'Topic :: Scientific/Engineering :: Mathematics',
        'Programming Language :: Python :: 3',
    ],
    license='Apache License Version 2.0',
    package_dir={
        '': 'src'
    },
    packages=find_namespace_packages(where='src'),
    namespace_packages=[
        "wai",
        "wai.annotations",
    ],
    version="1.0.0",
    author='Peter Reutemann',
    author_email='fracpete@waikato.ac.nz',
    install_requires=[
        "wai.annotations.core>=0.1.1",
    ],
    entry_points={
        "wai.annotations.plugins": [
            # sinks
            "label-dist-ic=wai.annotations.imgstats.format.labeldist.specifier:LabelDistributionICOutputFormatSpecifier",
            "label-dist-is=wai.annotations.imgstats.format.labeldist.specifier:LabelDistributionISOutputFormatSpecifier",
            "label-dist-od=wai.annotations.imgstats.format.labeldist.specifier:LabelDistributionODOutputFormatSpecifier",
            "to-annotation-overlay-od=wai.annotations.imgstats.format.annotation_overlay.specifier:AnnotationOverlayODOutputFormatSpecifier",
        ]
    }
)



from abc import ABC
from typing import Type, Tuple

from wai.annotations.core.component import Component
from wai.annotations.core.domain import DomainSpecifier
from wai.annotations.core.specifier import SinkStageSpecifier


class AnnotationOverlayODOutputFormatSpecifier(SinkStageSpecifier, ABC):
    """
    Base specifier for the label-dist in each known domain.
    """
    @classmethod
    def description(cls) -> str:
        return "Generates an image with all the annotation shapes (bbox or polygon) overlayed."

    @classmethod
    def components(cls) -> Tuple[Type[Component], ...]:
        from wai.annotations.imgstats.format.annotation_overlay.component import AnnotationOverlay
        return AnnotationOverlay,

    """
    Specifier for annotation-overlay-od in the object-detection domain.
    """
    @classmethod
    def domain(cls) -> Type[DomainSpecifier]:
        from wai.annotations.domain.image.object_detection import ImageObjectDetectionDomainSpecifier
        return ImageObjectDetectionDomainSpecifier
